//
//  ptfr.c
//  ptfr - partial tracking file resampler
//
//  Created by Stanko Juzbasic on 3/7/15.
//  Adapted from the code published as mentioned herewith:
//  Paper: CUBIC SPLINE INTERPOLATION: A REVIEW, by George Wolberg, 1988, CUCS
//         Technical Report CUCS-389-88

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

//__________________________________amp2db_________________________________
double
amp2db(double avalue)	//converts amplitudes to db
{
    if( avalue <= 0. ) return -140.;
    else avalue = 20.*log10(avalue);
    return avalue;
}
//__________________________________db2amp_________________________________
double
db2amp(double db)       //converts db to amplitudes
{
    if (db < -140.) db = -140;
	return(pow(10.,((db)/20.)));
}
//_________________________________________________________________________
/*
 CU CS George Wolberg, 1988
 Interpolating cubic spline function for irregularly-spaced points
 Input: Y1 is a list of irregular data points (len1 entries)
 Their x-coordinates are spec~ied in X1
 Output: Y2 <- cubic spline sampled according to X2 (len2 entries)
 Assume that X1,X2 entries are monotonically increasing
 */
//_________________________________________________________________________
/*
 Y0 <- Computed 1st derivative of data in X,Y (len entries) 
 The not-a-knot boundary condition is used
 */
//_________________________________________________________________________

void
tridiag_gen(A,B,C,D,len)
double *A, *B, *C, *D;
int len;
{
    int i;
    double b, *F;
    F = (double *) calloc(len, sizeof(double));
    /* Gauss elimination; forward substitution */
    b =B[0];
    D[0] = D[0] / b;
    for(i= 1;i<len;i++){
        F[i] = C[i-1] / b;
        b = B[i] - A[i]* F[i];
        if(b == 0) {
            fprintf(stderr,"getYD-gen: divide-by-zero");
            exit(1);
        }
        D[i] = (D[i] - D[i-1]*A[i]) / b;
    }
    /* backsubstitution */
    for(i = len-2; i >= 0; i--) D[i]-= (D[i+1]* F[i+1]);
                                  
free((char *) F);
return;
}
//_________________________________________________________________________

void
getYD_gen(X,Y,YD,len)
double *X, *Y, *YD;
int len;
{
int i;
double h0, h1, r0, r1, *A, *B, *C;
    // allocate memory for tridiagonal bands A,B,C
    A = (double *) calloc(len, sizeof(double));
    B = (double *) calloc(len, sizeof(double));
    C = (double *) calloc(len, sizeof(double));
    /* init first row data*/
    h0 = X[1]- X[0]; h1 = X[2]- X[1];
    r0 = (Y[1] - Y[0]) / h0; r1 = (Y[2]- Y[1]) / h1;
    B[0] = h1 * (h0+h1);
    C[0] = (h0+h1) * (h0+h1);
    YD[0] = r0*(3*h0*h1 + 2*h0*h1) + r1*h0*h0;
    /*init tridiagonal bands A, B, C, and column vector YD */
    /*YD will later be used to return the derivatives */
    for(i = 1; i < len-1; i++) {
    h0 =X[i]- X[i-1]; h1 =X[i+1] - X[i];
    r0 = (Y[i]- Y[i-1]) / h0; r1 = (Y[i+1]- Y[i]) / h1;
    A[i] = h1;
    B[i] = 2 * (h0+h1);
    C[i] = h0;
    YD[i] = 3 * (r0*h1 + r1*h0);
    }
    /*last row*/
    A[i] = (h0+h1)*(h0+h1);
    B[i] = h0*(h0+h1);
    YD[i] = r0*h1*h1 + r1*(3*h0*h1 + 2*h0*h0);
    /* solve for the tridiagonal matrix: YD=YD*inv(tridiag matrix) */
    tridiag_gen(A,B,C,YD,len);
    free((char *) A);
    free((char *) B);
    free((char *) C);
return;
}
//_________________________________________________________________________


void
ispline_gen(X1,Y1,len1,X2,Y2,len2)
double *X1, *Y1, *X2, *Y2;
int len1, len2;
{
    int i, j;
    double *YD, A0, A1, A2, A3, x, dx, dy, p1, p2, p3;
    /* compute 1st derivatives at each pOint -> YD */
    YD = (double*) calloc(len1, sizeof(double));
    getYD_gen(X1,Y1,YD,len1);
        /* error checking*/
        if(X2[0] < X1[0] || X2[len2-1] > X1 [len1-1]) {
            fprintf(stderr,"ispline_gen: Out of range\n");
            exit(1);
            }
        /*
        p1 is left endpoint of interval
        p2 is resampling position
        p3 is right endpoint of interval
        j is input index for current interval
        */
        p3 = X2[0] - 1;
        /*force coefficient initialization*/
        for(i=j=0; i < len2; i++) {
        /* check if in new interval */
        p2 = X2[i];
      
        if(p2 > p3) {
        /*for find the interval which contains p2 */
        for(; j<len1 && p2>X1[j]; j++);
        if(p2 < X1[j]) j--;
        p1 =X1[j]; /*updateleft endpoint*/
        p3 = X1[j+1]; /* update right endpoint */
    
        /* compute spline coefficients */
        dx =1.0/ (X1[j+1]- X1[j]);
        dy = (Y1[j+1]- Y1[j])*dx;
        A0 = Y1[j];
        A1 =YD[j];
        A2 = dx * (3.0*dy - 2.0*YD[j]- YD[j+1]);
        A3 = dx*dx * (-2.0*dy + YD[j] + YD[j+1]);
        }
        /* use Horner's rule to calculate cubic polynomial*/
        x =p2 - p1;
        Y2[i] = ((A3*x + A2)*x + A1)*x + A0;
        }
        free((char *) YD);
return;
}//_______________________________________________________________________

void
usage (void)
{
    printf("USAGE:\n");
    printf("      ptfr <infile> <dur>\n");
    printf("           <infile> is in LISP IRCAM format\n");
    printf("           <dur> = time spacing between resampled breakpoints\n");    
}
//________________________________________________________________________

int
main(int argc, const char * argv[])
{
    double **X1, **Y1, **X2, **Y2, **Z1, **Z2;
    double *len1;
    double *len2;
    double start, end, step;
    char text[256];
    FILE *afPtr, *sfPtr;
    char afname[256], *sfname;
    int partials;

    
    if(argc == 1){
    usage();
    exit(1);    
    }
    printf("============= RP ==============\n");
    strcpy (afname, argv[1]);
    
    if ((afPtr = fopen(afname, "r")) == NULL){
        printf("File %s not found\n", afname);
        exit(1);
    }
    
    if(argv[2] == NULL)step = 0.05;
    else step = atof(argv[2]);
    
    fscanf(afPtr,"%s",text);		// get & discard the initial "(" character    
    fscanf(afPtr, "%s", text);		// read the "PARTIALS" text    
    fscanf(afPtr, "%s", text);		// get the number of partials
    partials = (int)atoi(text);
    
    //parray = (Breakpoint**)malloc(partials*sizeof(Breakpoint*));//---- the array-------
    len1 = (double*)calloc(partials, sizeof(double));
    len2 = (double*)calloc(partials, sizeof(double));
    

    fscanf(afPtr,"%s",text); // discard the "(" text
    fscanf(afPtr,"%s",text); // discard the "POINTS" text
    fscanf(afPtr,"%s",text); // get the number of points for this partial
    
    rewind(afPtr);
    
    X1 = (double**)calloc((size_t)partials, sizeof(double*));
    Y1 = (double**)calloc((size_t)partials, sizeof(double*));
    Z1 = (double**)calloc((size_t)partials, sizeof(double*));
    X2 = (double**)calloc((size_t)partials, sizeof(double*));
    Y2 = (double**)calloc((size_t)partials, sizeof(double*));
    Z2 = (double**)calloc((size_t)partials, sizeof(double*));

    
    fscanf(afPtr,"%s",text);		// get & discard the initial "(" character
    fscanf(afPtr, "%s", text);		// read the "PARTIALS" text
    fscanf(afPtr, "%s", text);		// get the number of partials
    
    for (int x=0; x<partials; x++)
    {

        fscanf(afPtr,"%s",text); // discard the "(" text

        fscanf(afPtr,"%s",text); // discard the "POINTS" text
        fscanf(afPtr,"%s",text); // get the number of points for this partial
        len1[x] = atoi(text);
        //npnts[x] = points;
        //    parray[x] = (Breakpoint*)malloc((points)*sizeof(Breakpoint));
        // Allocate source arrays
        X1[x] = (double*)calloc((size_t)len1[x], sizeof(double));
        Y1[x] = (double*)calloc((size_t)len1[x], sizeof(double));
        Z1[x] = (double*)calloc((size_t)len1[x], sizeof(double));
        
        for (int y=0; y<(int)len1[x]; y++)
        {
            fscanf(afPtr,"%s",text);				//time
            
            //parray[x][y].time = (float)atof(text);
            X1[x][y] = atof(text);
            
            fscanf(afPtr,"%s",text);				//freq
            
            Y1[x][y] = atof(text);
            if(Y1[x][y] < 4.)Y1[x][y] = 4.;
            if(Y1[x][y] > 22050)Y1[x][y] = 22050.;
            fscanf(afPtr,"%s",text);			   //amplitude
            Z1[x][y]    = db2amp(atof(text));
            //nf = 0.;
            //if(parray[x][y].amp > *maxa)*maxa = parray[x][y].amp;
            
            
        }
        fscanf(afPtr,"%s",text); 				/* discard the ")" */
        //goto finish;
        /*
        if(strncmp(text, ")", 1) != 0)
            if(strncmp(text, "(", 1) == 0)
            {x++;
                goto audiosculpt_10_k_bug_fix;
            }
         */
    //Allocate destination arrays
        int l = (int)(len1[x] - 1);
        start = X1[x][0]; end = X1[x][l];
        l     = (int)((end- start)/step);
        len2[x]  = (double)l;
        printf("Len2: %d points\n",l);
        
        X2[x] = (double*)calloc((size_t)len2[x], sizeof(double));
        Y2[x] = (double*)calloc((size_t)len2[x], sizeof(double));
        Z2[x] = (double*)calloc((size_t)len2[x], sizeof(double));
        for(int j = 0; j < (int)len2[x]; j++)
            X2[x][j] = start + j*step;
        
    }
    fscanf(afPtr,"%s",text); 				/* discard the ")" */
    printf("\n");

    fclose(afPtr);

    for(int x = 0; x < partials; x++){
    
    ispline_gen(X1[x],Y1[x],len1[x],X2[x],Y2[x],len2[x]);
    

    ispline_gen(X1[x],Z1[x],len1[x],X2[x],Z2[x],len2[x]);
 

    int l = (int)len2[x];
    for(int j = 0; j < l; j++)
        printf("%lf --> %lf\t%lf\n", X2[x][j], Y2[x][j], Z2[x][j]);
    printf("\n");
    }
    

    
rename:
    printf("Enter file name: \n");
    printf("                 ");
    scanf("%s",text);
    sfname = calloc(strlen(text)+1, sizeof(char));
    strcpy(sfname, text);
    if ((sfPtr = fopen(sfname, "r")) != NULL){
        printf("File %s exists\n",sfname);
        fclose(sfPtr);
        free(sfname);
        goto rename;
    }
    
    if ((sfPtr = fopen(sfname, "w")) == NULL){
        printf("Cannot create file %s\n",sfname);
        goto fail;
    }
    
    fprintf(sfPtr,"( PARTIALS\t%d\n", partials);
    for(int p=0;p<partials;p++)
	{
    fprintf(sfPtr,"( POINTS\t%d\n",(int)len2[p]);
        for(int q=0; q<len2[p];q++)
        {
	 		fprintf(sfPtr,"%9.3f\t",(float)X2[p][q]);
			fprintf(sfPtr,"%9.3f\t",(float)Y2[p][q]);
	 		fprintf(sfPtr,"%9.3f",(float)amp2db(Z2[p][q]));
	 		fprintf(sfPtr,"\n");
        }
		fprintf(sfPtr,")\n");
	}
	fprintf(sfPtr,")\n");
    fclose (sfPtr);
    printf("Hello, World!\n");
    
fail:
    free(sfname);
    for(int x = 0; x < partials; x++){
        free(X1[x]); free(Y1[x]); free(Z1[x]);
        free(X2[x]); free(Y2[x]); free(Z2[x]);
    }
    free(len1); free(len2);
    free(X1); free(Y1); free(Z1);
    free(X2); free(Y2); free(Z2);
    return 0;
}

